--- A module for detecting the aircraft model.

local M = {}

--- Detects an aircraft model from an object.
--- Models are used to implement aircraft-specific sensor data or
--- code for issuing aircraft controls.
--- @return      string|nil  # model name (string) or nil for generic model
function M.detectAircraftModel()
    local model_name = v.config.model
    if model_name == "mtngoat" then
        return "mtngoat" -- <https://www.beamng.com/resources/mtn-goat.29759/> MTN-Goat Author: Shadow_Clan1965 
    elseif model_name == "me262" then
        return "me262" -- <https://www.beamng.com/resources/me262.2598/> Me262 Author: Stoat Muldoon
    end

    local mainpart_name = v.config.mainPartName
    if mainpart_name == "B25Mitchell" then
        return "B25Mitchell" -- <https://www.beamng.com/resources/b25-mitchell.2366/> B25 Mitchell Author: Joelvb
    end

    return nil  -- Generic model.
end

return M
