angular.module("av331NgWheel", []).directive("av331NgWheel", ["$parse", function($parse) {
    return function(scope, element, attr) {
        var fn = $parse(attr.av331NgWheel);
        element.bind("wheel", function (event) {
            scope.$apply(function () {
                fn(scope, { $event: event });
            });
        });
    };
}]);

angular.module('beamng.apps').directive('av331autopilot', [function () {
    /* const */ var GELUA_EXTENSION_ROOT = "extensions.av331autopilotg";
    /* const */ var EXTENSION_ROOT = "extensions.av331autopilot";
    /* const */ var KNOTS_IN_METERS_PER_SECOND = 1.94384449;
    /* const */ var FEET_PER_MINUTE_IN_METERS_PER_SECOND = 196.850394;
    /* const */ var FEET_IN_METERS = 3.2808399;

	/* const */ var settingsPath = "settings/ui_apps/aviator331/autopilot/settings.json";
    var appSettings = {};

    var findSeaLevelLua = `(function ()
        local water = scenetree.findObject("Ocean")
        if not water then
            local objects = scenetree.findClassObjects("WaterPlane")
            for _, name in pairs(objects) do
                local obj = scenetree.findObject(name)
                if string.find(name, "Ocean") then
                    water = obj
                    break
                end
            end
        end

        if water then
            return water.position:getColumn(3).z
        else
            return 0
        end
    end)()`;

    var bootstrapLua = `
        local autopilotExtensionName = "av331autopilot"
        if not extensions.isExtensionLoaded(autopilotExtensionName) then
            extensions.load(autopilotExtensionName)
        end
        extensions[autopilotExtensionName].setup()
    `;

    function isValidValue(value, minValue, maxValue) {
        return value != null && Number.isFinite(value) && value >= minValue && value <= maxValue;
    }

    function formatValue(value, digits, sign, maxValue) {
        var minValue = sign ? -maxValue : 0;
        if (value == null || value < minValue || value > maxValue) return "".padStart(digits + (sign ? 1 : 0), "-");
        var result = Math.abs(value).toString().padStart(digits, "0");
        if (sign) {
            if (value > 0) sign = "+";
            else if (value < 0) sign = "-";
            else sign = "\u00a0";
            result = sign + result;
        }
        return result;
    }

    function clamp(v, vMin, vMax) {
        return Math.max(vMin, Math.min(vMax, v));
    }

    var seaLevel = 0;

    return {
        templateUrl: '/ui/modules/apps/Av331Autopilot/app.html',
        replace: false,
        restrict: 'E',
        scope: false,
        controller: function ($scope, $element, $attrs) {
            var vm = this;

            function resetDefaults() {
                vm.enabled = false;
                vm.autothrottle = false;
                vm.mode = "";
                vm.alt = null;
                vm.vs = null;
                vm.spd = null;
                vm.hdg = null;
                vm.liveAlt = 0;
                vm.liveVs = 0;
                vm.liveSpd = 0;
                vm.liveHdg = 0;
                vm.spdOn = false;
                vm.hdgOn = false;
                vm.displayAlt = "-----";
                vm.displayVs = "------";
                vm.displayHdg = "---";
                vm.displaySpd = "---";
            }
            resetDefaults();

            var paused = false;
            var inReplay = false;
            var switched = false;

            var streamsList = ['av331autopilotstream'];

            StreamsManager.add(streamsList);

            /* bngApi.engineLua('jsonReadFile(' + bngApi.serializeToLua(settingsPath) + ')', (settings) => {
                appSettings = settings;
            }); */
            
            function isInReplay(replayState) {
                return replayState && replayState.state != "inactive" && replayState.state != "recording";
            }

            $scope.$on("$destroy", function () {
                // bngApi.engineLua('jsonWriteFile(' + bngApi.serializeToLua(settingsPath) + ', ' + bngApi.serializeToLua(appSettings) + ', ' + bngApi.serializeToLua(true) + ')');
                StreamsManager.remove(streamsList);
            });
            $scope.$on("physicsStateChanged", function (event, running) {
                paused = !running;
            });
            $scope.$on("replayStateChanged", function (event, val) {
                inReplay = isInReplay(val);
            });
            $scope.$emit('requestPhysicsState');
            $scope.$on("VehicleReset", function (_, data) {
                switched = true;
                resetDefaults();
            });
            $scope.$on("VehicleChange", function (_, data) {
                switched = true;
                resetDefaults();
            });
            $scope.$on("VehicleFocusChanged", function (_, data) {
                switched = true;
                resetDefaults();
            });

            // initial pause and replay state
            bngApi.engineLua("simTimeAuthority.getPause()",
                (isPaused) => {
                    paused = isPaused;
                });
            $scope.$emit('requestPhysicsState');
            bngApi.engineLua(`
                (function ()
                    if extensions.isExtensionLoaded("core_replay") then
                        return extensions.core_replay.state
                    else
                        return nil
                    end
                end)()`,
                (replayState) => {
                    inReplay = isInReplay(replayState);
                });
            bngApi.engineLua("(function () if be:getFileStream() then be:getFileStream():requestState() end end)()");

            function updateDisplays() {
                vm.displayAlt = formatValue(vm.alt, 5, false, 99999);
                vm.displayVs = formatValue(vm.vs, 5, true, 99999);
                vm.displayHdg = formatValue(vm.hdg, 3, false, 360);
                vm.displaySpd = formatValue(vm.spd, 3, false, 999);
            }
            var lastLuaData = null;
            function updateDataFromLuaCached(luaData) {
                vm.enabled = luaData.on;
                vm.autothrottle = luaData.athr;
                vm.mode = luaData.mode || "";
                vm.spdOn = luaData.spdOn;
                vm.hdgOn = luaData.hdgOn;
                vm.vs = luaData.vs == null ? null : Math.round(luaData.vs * FEET_PER_MINUTE_IN_METERS_PER_SECOND);
                vm.alt = luaData.alt == null ? null : Math.round((luaData.alt - seaLevel) * FEET_IN_METERS);
                vm.spd = luaData.spd == null ? null : Math.round(luaData.spd * KNOTS_IN_METERS_PER_SECOND);
                vm.hdg = luaData.hdg == null ? null : Math.round(luaData.hdg);
                updateDisplays();
            }
            function updateDataFromLua(luaData) {
                lastLuaData = luaData;
                updateDataFromLuaCached(luaData);
            }

            function setMode(mode) {
                if (mode)
                    bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("mode",${bngApi.serializeToLua(mode)})`);
                else
                    bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("mode",nil)`);
                // only update viewmodel if paused - otherwise it 'flickers'
                if (paused) vm.mode = mode;
            }
            function setAlt(alt) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("alt",${bngApi.serializeToLua(alt / FEET_IN_METERS + seaLevel)})`);
                // only update viewmodel if paused - otherwise it 'flickers'
                if (paused) {
                    vm.alt = alt;
                    updateDisplays();
                }
            }
            function setVs(vs) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("vs",${bngApi.serializeToLua(vs / FEET_PER_MINUTE_IN_METERS_PER_SECOND)})`);
                if (paused) {
                    vm.vs = vs;
                    updateDisplays();
                }
            }
            function setSpd(spd) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("spd",${bngApi.serializeToLua(spd / KNOTS_IN_METERS_PER_SECOND)})`);
                if (paused) {
                    vm.spd = spd;
                    updateDisplays();
                }
            }
            function setHdg(hdg) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("hdg",${bngApi.serializeToLua(hdg)})`);
                if (paused) {
                    vm.hdg = hdg;
                    updateDisplays();
                }
            }
            function setEnabled(enabled) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("on",${bngApi.serializeToLua(enabled)})`);
                // only update viewmodel if paused - otherwise it 'flickers'
                if (paused) vm.enabled = enabled;
            }
            function setAutothrottle(autothrottle) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("athr",${bngApi.serializeToLua(autothrottle)})`);
                // only update viewmodel if paused - otherwise it 'flickers'
                if (paused) vm.autothrottle = autothrottle;
            }
            function setSpdOn(spdOn) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("spdOn",${bngApi.serializeToLua(spdOn)})`);
                // only update viewmodel if paused - otherwise it 'flickers'
                if (paused) vm.spdOn = spdOn;
            }
            function setHdgOn(hdgOn) {
                bngApi.activeObjectLua(`${bootstrapLua}; ${EXTENSION_ROOT}.setAutopilotSetting("hdgOn",${bngApi.serializeToLua(hdgOn)})`);
                // only update viewmodel if paused - otherwise it 'flickers'
                if (paused) vm.hdgOn = hdgOn;
            }

            $element.ready(function () {
                parent = $element[0];
                updateDisplays();
                bngApi.engineLua(findSeaLevelLua,
                    function (newSeaLevel) {
                        seaLevel = newSeaLevel;
                        if (lastLuaData) updateDataFromLuaCached(lastLuaData);
                    });
            });

            $scope.$on("av331AutopilotChange", function (_, luaData) {
                if (luaData) updateDataFromLua(luaData);
            });
            $scope.$on("av331SeaLevelChange", function (_, newSeaLevel) {
                seaLevel = newSeaLevel;
                if (lastLuaData) updateDataFromLuaCached(lastLuaData);
            });
            
            vm.toggleMode = function (mode) {
                setMode(vm.mode == mode ? null : mode);
            }
            vm.setAlt = function () {
                if (!isValidValue(vm.liveAlt, 0, 99999)) return;
                setAlt(vm.liveAlt);
            }
            vm.setVs = function () {
                if (!isValidValue(vm.liveVs, -99999, 99999)) return;
                setVs(vm.liveVs);
            }
            vm.setSpd = function () {
                if (!isValidValue(vm.liveSpd, 0, 999)) return;
                setSpd(vm.liveSpd);
            }
            vm.setHdg = function () {
                if (!isValidValue(vm.liveHdg, 0, 360)) return;
                setHdg(vm.liveHdg);
            }
            vm.toggleEnabled = function () {
                setEnabled(!vm.enabled);
            }
            vm.toggleAutothrottle = function () {
                setAutothrottle(!vm.autothrottle);
            }
            vm.toggleSpdOn = function () {
                setSpdOn(!vm.spdOn);
            }
            vm.toggleHdgOn = function () {
                setHdgOn(!vm.hdgOn);
            }

            vm.copyAlt = function () {
                if (vm.alt != null && vm.alt >= 0 && vm.alt <= 99999) vm.liveAlt = vm.alt;
            }
            vm.copyHdg = function () {
                if (vm.hdg != null && vm.hdg >= 0 && vm.hdg <= 360) vm.liveHdg = vm.hdg;
            }
            vm.copySpd = function () {
                if (vm.spd != null && vm.spd >= 0 && vm.spd <= 999) vm.liveSpd = vm.spd;
            }
            vm.copyVs = function () {
                if (vm.vs != null && vm.vs >= -99999 && vm.vs <= 99999) vm.liveVs = vm.vs;
            }

            function getMouseWheelShift(wheelEvent) {
                var notch = wheelEvent.deltaMode == 0 ? 120 : 1;
                return -Math.round(wheelEvent.deltaY / notch) * (wheelEvent.shiftKey ? 10 : 1);
            }

            vm.adjustAltMouseWheel = function (wheelEvent) {
                var shift = getMouseWheelShift(wheelEvent);
                if (vm.alt != null && shift != 0) setAlt(vm.alt = clamp(vm.alt + shift * 100, 0, 99999));
            }
            vm.adjustHdgMouseWheel = function (wheelEvent) {
                var shift = getMouseWheelShift(wheelEvent);
                if (vm.hdg != null && shift != 0) {
                    var newHdg = vm.hdg + shift;
                    newHdg = (newHdg % 360 + 360) % 360;
                    setHdg(newHdg);
                    vm.hdg = newHdg;
                }
            }
            vm.adjustSpdMouseWheel = function (wheelEvent) {
                var shift = getMouseWheelShift(wheelEvent);
                if (vm.spd != null && shift != 0) setSpd(vm.spd = clamp(vm.spd + shift, 0, 999));
            }
            vm.adjustVsMouseWheel = function (wheelEvent) {
                var shift = getMouseWheelShift(wheelEvent);
                if (vm.vs != null && shift != 0) setVs(vm.vs = clamp(vm.vs + shift * 100, -99999, 99999));
            }
            
            $scope.$on('streamsUpdate', function (event, streams) {
                if (paused && !inReplay && !switched) return;
                var autopilotData = streams.av331autopilotstream;
                if (autopilotData) updateDataFromLua(autopilotData);
                switched = false;
            });
        },
        controllerAs: 'ctrl'
    };
}])